/*
 * Altogether: Xerox Alto microcode-level simulator
 * Main CPU simulation
 * $Id: cpu.c 93 2003-10-24 02:04:52Z eric $
 * Copyright 2001 Eric Smith <eric@brouhaha.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.  Note that permission is
 * not granted to redistribute this program under the terms of any
 * other version of the General Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111  USA
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "alto.h"
#include "debug.h"
#include "scheduler.h"
#include "cpu.h"
#include "main_mem.h"
#include "emulator.h"  /* we need the skip flag */


#define UCODE_PAGE_SIZE 1024

#define UCODE_ROM_SIZE (UCODE_ROM_PAGES * UCODE_PAGE_SIZE)
#define UCODE_RAM_SIZE (UCODE_RAM_PAGES * UCODE_PAGE_SIZE)
#define UCODE_RAM_BASE (UCODE_ROM_SIZE)
#define UCODE_SIZE (UCODE_ROM_SIZE + UCODE_RAM_SIZE)

uint32_t ucode [UCODE_SIZE];
int ubreak [UCODE_SIZE];


uint16_t constant_rom [CONSTANT_ROM_SIZE];

uint16_t l, t, cram_addr;
int l_alu_carry;

int rsel;  /* from rsel field of microinstruction, but may be
	      modified by emulator acsource and acdest uops */
uint16_t r [32];
uint16_t s [S_REG_BANKS][32];


uint16_t reset_mode_register;  /* bit 1<<n set if task n starts in ROM */

char *task_name [] =
{
  "emu", "t01", "t02", "t03", "dsc", "t05", "t06", "eth",
  "mrt", "dwt", "cur", "dht", "dvt", "par", "dwd", "t17"
};

int current_task, next_task, next2_task;
uint16_t task_request;  /* bit 1<<n set if task n requesting service */
uint16_t task_autoblock;  /* bit 1<<n set if task n autoblocks */

bool task_switch;  /* set when task switch happens */


int rdram_flag;  /* set by rdram, action happens on next cycle */
int wrtram_flag;  /* set by wrtram, action happens on next cycle */

uint16_t address_modifier;
uint16_t upc [TASK_COUNT];
int s_reg_bank [TASK_COUNT];

int normal_ram_bank [TASK_COUNT];
int extended_ram_bank [TASK_COUNT];

uint16_t current_upc;  /* adddress of microinstruction currently being executed,
		   * used only for trace output, etc.
		   */
uinst_t current_uinst;

uint16_t bus;  /* wire-AND */

uint16_t alu_out;
bool alu_carry_out;

uint16_t shifter_out;
bool shifter_carry_out;


int illegal_instruction;


dispatch_fn_t bus_select_fn [TASK_COUNT][16];
dispatch_fn_t f1_fn [TASK_COUNT][16];
dispatch_fn_t f2_fn [TASK_COUNT][16];



#define UCODE_INVERTED_BITS 0x00088400


static char *ucode_rom_filenames [16] =
{
  "55x.3", "64x.3", "65x.3", "63x.3", "53x.3", "60x.3", "61x.3", "62x.3",
  "54x.3", "74x.3", "75x.3", "73x.3", "52x.3", "70x.3", "71x.3", "72x.3"
};

void load_ucode_rom (void)
{
  int page, segment, addr;
  char *fn;
  FILE *f;
  unsigned char buf [UCODE_PAGE_SIZE];

  for (page = 0; page < UCODE_ROM_PAGES; page++)
    {
      for (addr = 0; addr < UCODE_PAGE_SIZE; addr++)
	{
	  ucode [(page << 10) + addr] = 0;  /* should this be all-ones? */
	}
      for (segment = 0; segment <= 7; segment++)
	{
	  fn = ucode_rom_filenames [(page << 3) + segment];
	  f = fopen (fn, "rb");
	  if (! f)
	    {
	      fprintf (stderr, "fatal - can't open ucode file %s\n", fn);
	      exit (3);
	    }
	  if (fread (buf, 1, UCODE_PAGE_SIZE, f) != UCODE_PAGE_SIZE)
	    {
	      fprintf (stderr, "fatal - error reading ucode file %s\n", fn);
	      exit (3);
	    }
	  for (addr = 0; addr < UCODE_PAGE_SIZE; addr++)
	    {
	      /* addresses are complemented */
	      ucode [(page << 10) + addr] |=
		(buf [(UCODE_PAGE_SIZE - 1) - addr] << ((7 - segment) * 4));
	    }
	  fclose (f);
	}
      for (addr = 0; addr < UCODE_PAGE_SIZE; addr++)
	{
	  ucode [(page << 10) + addr] ^= ~ UCODE_INVERTED_BITS;
#if 0
	  printf ("%05o: %08x\n", (page << 10) + addr, ucode [(page << 10) + addr]);
#endif
	}
    }
}


bool check_ubreak (void)
{
  return (ubreak [upc [current_task]]);
}


void set_ubreak (uint16_t addr, bool value)
{
  ubreak [addr] = value;
}


static char *const_rom_filenames [4] =
{
  "c0.3", "c1.3", "c2.3", "c3.3"
};

static int unscramble_constant_address (int a)
{
  return (((a & 0x80) >> 7) |
	  ((a & 0x78) << 1) |
	  ((a & 0x04) >> 1) |
	  ((a & 0x02) << 1) |
	  ((a & 0x01) << 3));
}

static int unscramble_constant_data (int d)
{
  int i;
  int d2 = 0;
  for (i = 0; i <= 7; i++)
    d2 |= ((d & (1 << i)) << (15 - 2 * i));
  for (i = 8; i <= 15; i++)
    d2 |= ((d & (1 << i)) >> ((i - 8) * 2 + 1));
  return (d2);
}

void load_const_rom (void)
{
  int segment, addr;
  char *fn;
  FILE *f;
  unsigned char buf [CONSTANT_ROM_SIZE];

  for (addr = 0; addr < CONSTANT_ROM_SIZE; addr++)
    constant_rom [addr] = 0;
  for (segment = 0; segment < 4; segment++)
    {
      fn = const_rom_filenames [segment];
      f = fopen (fn, "rb");
      if (! f)
	{
	  fprintf (stderr, "fatal - can't open constants file %s\n", fn);
	  exit (3);
	}
      if (fread (buf, 1, CONSTANT_ROM_SIZE, f) != CONSTANT_ROM_SIZE)
	{
	  fprintf (stderr, "fatal - error reading constants file %s\n", fn);
	  exit (3);
	}
      for (addr = 0; addr < CONSTANT_ROM_SIZE; addr++)
	{
	  constant_rom [unscramble_constant_address (addr)] |=
	      (buf [addr] ^017) << (4 * segment);
	}
      fclose (f);
    }
  for (addr = 0; addr < CONSTANT_ROM_SIZE; addr++)
    {
      constant_rom [addr] = unscramble_constant_data (constant_rom [addr]);
    }
}


uint16_t bank_register_read (uint16_t address)
{
  return (0177760 |
	  (normal_ram_bank [address & 017] << 2) |
	  extended_ram_bank [address & 017]);
}


void bank_register_write (uint16_t address, uint16_t data)
{
  normal_ram_bank [address & 017] = (data >> 2) & 3;
  extended_ram_bank [address & 017] = data & 3;
}


void init_cpu_mmio (void)
{
  install_memory_mapped_io (0177740, 0177757,
			    bank_register_read, bank_register_write);
}


void soft_reset (void)
{
  int i;

  current_task = 0;
  for (i = 0; i < TASK_COUNT; i++)
    {
      upc [i] = i;
      if (! (reset_mode_register & (1 << i)))
	upc [i] += UCODE_RAM_BASE;
      s_reg_bank [i] = 0;
    }

  task_request = 0;  /* only emulator requesting */
  set_task_request (task_emulator);

  reset_mode_register = 0xffff;

  address_modifier = 0;
  rdram_flag = 0;
  wrtram_flag = 0;
}


void hard_reset (void)
{
  reset_mode_register = 0xffff;
  soft_reset ();
}


void nop (int task, int code, int after)
{
}

void bs_fn_bad (int task, int code, int after)
{
  if (! after)
    {
      illegal_instruction = 1;
      printf ("bad bus source field %d\n", code);
    }
}

void bs_fn_read_r (int task, int code, int after)
{
  char regname [4] = { 'r', '0', '0', '\0' };
  if (! after)
    {
      regname [1] += rsel >> 3;
      regname [2] += rsel & 7;
      drive_bus (regname, r [rsel]);
    }
}

void bs_fn_load_r (int task, int code, int after)
{
  if (! after)
    {
      drive_bus ("load_r", 0);
    }
  else
    {
      r [rsel] = shifter_out;
      tprintf (1, "r%02o <= %06o\n", rsel, shifter_out);
    }
}



void bs_fn_mem_data (int task, int code, int after)
{
  if (! after)
    drive_bus ("read_mem", read_main_mem ());
}


void bs_fn_mouse (int task, int code, int after)
{
  if (! after)
    {
      /* ??? */
    }
}


void f1_fn_bad (int task, int code, int after)
{
  if (! after)
    {
      illegal_instruction = 1;
      printf ("bad f1 field %d\n", code);
    }
}


void f1_fn_load_mar (int task, int code, int after)
{
  int bank;
  if (after)
    {
      if (current_uinst.f2 == f2_load_md)
	bank = extended_ram_bank [current_task];
      else
	bank = normal_ram_bank [current_task];
      load_main_mem_addr ((bank << 16) | alu_out);
    }
}


void f1_fn_task (int task, int code, int after)
{
  int i;
  int mask;
  if (! after)
    {
      for (i = 15, mask = 0100000; i >= 0; i--, mask >>= 1)
	if (task_request & mask)
	  {
	    next2_task = i;
	    tprintf (1, "task switch to %s\n", task_name [next2_task]);
	    if (task_autoblock & (1 << i))
	      clear_task_request (i);
	    if (next2_task != next_task)
	      task_switch = true;
	    return;
	  }
      fprintf (stderr, "fatal error - no tasks requesting service\n");
      exit (3);
    }
}


void f1_fn_block (int task, int code, int after)
{
  if (! after)
    {
      clear_task_request (task);
    }
}


void f2_fn_bad (int task, int code, int after)
{
  if (! after)
    {
      illegal_instruction = 1;
      printf ("bad f2 field %d\n", code);
    }
}

void f2_fn_bus_eq_0 (int task, int code, int after)
{
  if (after)
    {
      branch ("bus_eq_0", bus == 0);
    }
}

void f2_fn_shifter_lt_0 (int task, int code, int after)
{
  if (after)
    {
      branch ("shifter_lt_0", (shifter_out & 0100000) != 0);
    }
}

void f2_fn_shifter_eq_0 (int task, int code, int after)
{
  if (after)
    {
      branch ("shifter_eq_0", shifter_out == 0);
    }
}

void f2_fn_bus (int task, int code, int after)
{
  if (after)
    {
      branch ("bus", bus & 01777);
    }
}

void f2_fn_alucy (int task, int code, int after)
{
  if (after)
    {
      branch ("alucy", l_alu_carry);
    }
}

void f2_fn_load_md (int task, int code, int after)
{
  if (after)
    {
      if (current_uinst.f1 == f1_load_mar)
	; /* it's just part of an XMAR */
      else
	write_main_mem (bus);
    }
}




void install_bus_select_fn (int task, int code, dispatch_fn_t fn)
{
  bus_select_fn [task][code] = fn;
}

void install_f1_fn (int task, int code, dispatch_fn_t fn)
{
  f1_fn [task][code] = fn;
}

void install_f2_fn (int task, int code, dispatch_fn_t fn)
{
  f2_fn [task][code] = fn;
}


void uninstall_bus_select_fn (int task, int code)
{
  bus_select_fn [task][code] = bs_fn_bad;
}

void uninstall_f1_fn (int task, int code)
{
  f1_fn [task][code] = f1_fn_bad;
}

void uninstall_f2_fn (int task, int code)
{
  f2_fn [task][code] = f2_fn_bad;
}


void init_function_dispatch (void)
{
  int task, i;

  for (task = 0; task < TASK_COUNT; task++)
    {
      for (i = 0; i < 8; i++)
	{
	  install_bus_select_fn (task, i, & bs_fn_bad);
	}

      install_bus_select_fn (task, bs_read_r, & bs_fn_read_r);
      install_bus_select_fn (task, bs_load_r, & bs_fn_load_r);
      install_bus_select_fn (task, bs_no_source, & nop);
      install_bus_select_fn (task, bs_mem_data, & bs_fn_mem_data);
      install_bus_select_fn (task, bs_mouse, & bs_fn_mouse);

      for (i = 0; i < 16; i++)
	{
	  install_f1_fn (task, i, & f1_fn_bad);
	  install_f2_fn (task, i, & f2_fn_bad);
	}

      install_f1_fn (task, f1_nop,          & nop);
      install_f1_fn (task, f1_load_mar,     & f1_fn_load_mar);
      install_f1_fn (task, f1_task,         & f1_fn_task);
#if 0
      install_f1_fn (task, f1_block,        & f1_fn_block);
      /* don't install f1_block here - it's not appropriate for all tasks */
#endif
      install_f1_fn (task, f1_l_lsh_1,      & nop);  /* handled elsewhere */
      install_f1_fn (task, f1_l_rsh_1,      & nop);  /* handled elsewhere */
      install_f1_fn (task, f1_l_lcy_8,      & nop);  /* handled elsewhere */
      install_f1_fn (task, f1_constant,     & nop);  /* handled elsewhere */

      install_f2_fn (task, f2_nop,          & nop);
      install_f2_fn (task, f2_bus_eq_0,     & f2_fn_bus_eq_0);
      install_f2_fn (task, f2_shifter_lt_0, & f2_fn_shifter_lt_0);
      install_f2_fn (task, f2_shifter_eq_0, & f2_fn_shifter_eq_0);
      install_f2_fn (task, f2_bus,          & f2_fn_bus);
      install_f2_fn (task, f2_alucy,        & f2_fn_alucy);
      install_f2_fn (task, f2_load_md,      & f2_fn_load_md);
      install_f2_fn (task, f2_constant,     & nop);  /* handled elsewhere */
    }
}


void print_uinst (uinst_t *uinst)
{
  printf ("rsel %02o, aluf %02o, bs: %o, f1: %02o, f2: %02o, t %d, l %d, next %04o\n",
	  uinst->rsel,
	  uinst->aluf,
	  uinst->bs,
	  uinst->f1,
	  uinst->f2,
	  uinst->t,
	  uinst->l,
	  uinst->next);
}


static void alu (void)
{
  uint32_t result;
  switch (current_uinst.aluf)
    {
    case aluf_bus__alut:
      result = bus;
      break;
    case aluf_treg:
      result = t;
      break;
    case aluf_bus_or_t__alut:
      result = bus | t;
      break;
    case aluf_bus_and_t:
      result = bus & t;
      break;
    case aluf_bus_xor_t:
      result = bus ^ t;
      break;
    case aluf_bus_plus_1__alut:
      result = bus + 1;
      break;
    case aluf_bus_minus_1__alut:
      /* result = bus - 1; */
      result = bus + 0177777;
      break;
    case aluf_bus_plus_t:
      result = bus + t;
      break;
    case aluf_bus_minus_t:
      /* result = bus - t; */
      result = bus + (t ^ 0177777) + 1;
      break;
    case aluf_bus_minus_t_minus_1:
      /* result = (bus - t) - 1; */
      result = bus + (t ^ 0177777);
      break;
    case aluf_bus_plus_t_plus_1__alut:
      result = bus + t + 1;
      break;
    case aluf_bus_plus_skip__alut:
      result = bus + skip;
      break;
    case aluf_bus_and_t__alut:
      result = bus & t;
      break;
    case aluf_bus_and_not_t:
      result = bus & ~ t;
      break;
    case aluf_zero:
      result = 0;
      break;
    case aluf_undef:
      illegal_instruction = 1;
      result = 0;
    }
  alu_out = (result & 0177777);
  alu_carry_out = ((result & 0200000) != 0);
}

static uint16_t shifter (void)
{
  uint16_t result;
  int shifter_function;
  int magic, dns;

  magic = (current_task == task_emulator) && (current_uinst.f2 == f2_em_magic);
  dns   = (current_task == task_emulator) && (current_uinst.f2 == f2_em_dns);

  if (dns)
    shifter_function = (ir >> 6) & 2;
  else
    switch (current_uinst.f1)
      {
      case f1_l_lsh_1:	shifter_function = 1;  break;
      case f1_l_rsh_1:	shifter_function = 2;  break;
      case f1_l_lcy_8:	shifter_function = 3;  break;
      default:          shifter_function = 0;  break;
      }

  switch (shifter_function)
    {
    case 0:
      shifter_carry_out = l_alu_carry;
      result = l;
      break;
    case 1:
      shifter_carry_out = l >> 15;
      result = (l << 1) & 0177776;
      if (magic && (t & 0100000))
	result |= 1;
      else if (dns)
	result |= l_alu_carry;
      break;
    case 2:
      shifter_carry_out = l & 1;
      result = (l >> 1);
      if (magic && (t & 1))
	result |= 0100000;
      else if (dns)
	result |= (l_alu_carry << 15);
      break;
    case 3:
      shifter_carry_out = l_alu_carry;
      result = (l >> 8) | ((l & 0377) << 8);
      break;
    }
  return (result);
}


static void rdram (void)
{
  int addr;
  uint32_t val;

  if (cram_addr & 004000)
    {
      /* read ROM 0 at current upc */
      addr = current_upc & 01777;
    }
  else
    {
      /* read RAM at cram_addr */
      addr = (cram_addr & 001777) +
	UCODE_RAM_BASE + ((cram_addr >> 2) & 006000);
    }
  if (cram_addr & 002000)
    drive_bus ("rdram", (val >> 16) ^ (UCODE_INVERTED_BITS >> 16));
  else
    drive_bus ("rdram", (val & 0177777) ^ (UCODE_INVERTED_BITS & 0xffff));
  rdram_flag = 0;
}


static void wrtram (void)
{
  int addr;
  addr = (cram_addr & 001777) +
    UCODE_RAM_BASE + ((cram_addr >> 2) & 006000);
  ucode [addr] = (((l ^ (UCODE_INVERTED_BITS >> 16)) << 16) |
		  (alu_out ^ (UCODE_INVERTED_BITS & 0xffff)));
  wrtram_flag = 0;
}


void start_rdram (void)
{
  rdram_flag = 1;
}


void start_wrtram (void)
{
  wrtram_flag = 1;
}


static int t_source [16] =
{
  1, 0, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0, 0
};

#define bit_mask(bits) ((1 << bits) - 1)


void fetch_uinst (uint16_t addr, uinst_t *uinst)
{
  uint32_t val;

  val = ucode [addr];
  uinst->rsel = (val >> 27) & bit_mask (5);
  uinst->aluf = (val >> 23) & bit_mask (4);
  uinst->bs   = (val >> 20) & bit_mask (3);
  uinst->f1   = (val >> 16) & bit_mask (4);
  uinst->f2   = (val >> 12) & bit_mask (4);
  uinst->t    = (val >> 11) & bit_mask (1);
  uinst->l    = (val >> 10) & bit_mask (1);
  uinst->next = (val >>  0) & bit_mask (10);
}


void pre_instruction_trace (void)
{
  uinst_t uinst;

  printf ("time %llu, %s-%04o: ",
	  current_time,
	  task_name [next_task],
	  upc [next_task]);
  fetch_uinst (upc [next_task], & uinst);
  print_uinst (& uinst);
}


void cpu_cycle (void)
{
  int do_bs;

  /* possible task switching */
  current_task = next_task;
  next_task = next2_task;

  /* "fetch" microinstruction */
  current_upc = upc [current_task];

  fetch_uinst (current_upc, & current_uinst);

  rsel = current_uinst.rsel;
#if 0
  if (current_trace_level >= 1)
    print_uinst ();
#endif

  do_bs = ((current_uinst.f1 != f1_constant) &&
	   (current_uinst.f2 != f2_constant));

  if (do_bs && (current_uinst.bs == bs_mem_data) &&
      check_read_main_mem_stall ())
    {
      tprintf (1, "memory read stall\n");
      return;
    }

  if ((current_uinst.f2 == f2_load_md) &&
      (current_uinst.f1 != f1_load_mar) &&
      check_write_main_mem_stall ())
    {
      tprintf (1, "memory write stall\n");
      return;
    }

  /*
   * point upc to next instruction, with any address modification from
   * previous instruction
   */
  upc [current_task] = (current_upc & ~ 01777) |
    current_uinst.next | address_modifier;

  address_modifier = 0;  /* no address modifier yet for this instruction */
  bus = 0xffff;  /* precharge the wire-AND data bus */
  illegal_instruction = 0;  /* assume legal */

  if (rdram_flag)
    rdram ();

  do_bs = ((current_uinst.f1 != f1_constant) &&
	   (current_uinst.f2 != f2_constant));

  if ((current_uinst.f1 == f1_constant) ||
      (current_uinst.f2 == f2_constant) ||
      (current_uinst.bs >= 4))
    drive_bus ("constant", constant_rom [(current_uinst.rsel << 3) + current_uinst.bs]);

  /* pre-f2 has to be done before pre-bs because emulator f2 acsource or
     acdest may change rsel */
  (* f2_fn [current_task][current_uinst.f2]) (current_task, current_uinst.f2, 0);

  if (do_bs)
    (* bus_select_fn [current_task][current_uinst.bs]) (current_task, current_uinst.bs, 0);

  (* f1_fn [current_task][current_uinst.f1]) (current_task, current_uinst.f1, 0);

  alu ();
  shifter_out = shifter ();

  if (do_bs)
    (* bus_select_fn [current_task][current_uinst.bs]) (current_task, current_uinst.bs, 1);

  (* f1_fn [current_task][current_uinst.f1]) (current_task, current_uinst.f1, 1);
  (* f2_fn [current_task][current_uinst.f2]) (current_task, current_uinst.f2, 1);

  /* wrtram uses the last L register, must do this before updating L */
  if (wrtram_flag)
    wrtram ();

  if (current_uinst.l)
    {
      l = alu_out;
      l_alu_carry = alu_carry_out;
      tprintf (1, "alucy, l <= %o, %06o\n", l_alu_carry, l);
    }

  if (current_uinst.t)
    {
      if (t_source [current_uinst.aluf])
	t = alu_out;
      else
	t = bus;
      cram_addr = alu_out;
      tprintf (1, "t <= %06o\n", t);
    }
}


void set_task_request (int task)
{
  task_request |= (1 << task);
}

void clear_task_request (int task)
{
  task_request &= ~ (1 << task);
}

void set_task_autoblock (int task)
{
  task_autoblock |= (1 << task);
}

